from banterbot.characters.android import run as android
from banterbot.characters.bartender import run as bartender
from banterbot.characters.chef import run as chef
from banterbot.characters.historian import run as historian
from banterbot.characters.quiz import run as quiz
from banterbot.characters.teacher_french import run as teacher_french
from banterbot.characters.teacher_mandarin import run as teacher_mandarin
from banterbot.characters.therapist import run as therapist

__all__ = ["android", "bartender", "chef", "historian", "quiz", "teacher_french", "teacher_mandarin", "therapist"]
